"""module image_s3_batch. Represents
a batch of images on S3

"""
import os
from typing import List, Any, Union, Callable
from navalmartin_mir_aws_utils.aws_credentials import AWSCredentials_S3
from navalmartin_mir_aws_utils.boto3_client import get_aws_s3_client
from navalmartin_mir_aws_utils.s3_utils import (
    expand_s3_iterator_contents,
    expand_s3_iterator_common_prefixes,
)

S3_URI = "s3://"


class FilePathBatch(object):
    def __init__(
        self,
        s3_credentials: AWSCredentials_S3,
        delimiter: str = "/",
        do_build_client: bool = True,
    ):
        self.aws_bucket_credentials = s3_credentials
        self.files: List[Any] = []
        self.delimiter = delimiter
        self._current_pos: int = -1
        self._client: Any = None

        if do_build_client:
            self.build_client()

    def __len__(self):
        return len(self.files)

    def __iter__(self):
        self._current_pos = 0
        return self

    def __next__(self):
        if len(self.files) == 0:
            raise StopIteration

        if self._current_pos < len(self.files):
            result = self.files[self._current_pos]
            self._current_pos += 1
            return result
        else:
            self._current_pos = -1
            raise StopIteration

    def __getitem__(self, key: Union[int, str]) -> str:
        """Returns the image that corresponds to the given key

        Parameters
        ----------
        key

        Returns
        -------

        A string representing the image in the batch
        """
        if type(key) == int:
            if key >= len(self.files):
                raise ValueError(
                    f"Invalid key. Integer key {key} cannot be >= {len(self.files)}"
                )

            key = self.files[key]
            return key
        elif type(key) == str:
            if key not in self.files:
                raise ValueError(f"key={key} not in {self.files}")
            return key
        else:
            raise ValueError(f"key type {type(key)} is not valid")

    def reinit(
        self,
        s3_credentials: AWSCredentials_S3,
        delimiter: str = "/",
        do_build_client: bool = True,
    ):
        """Set the file batch to the state just after the constructor
        was called

        Parameters
        ----------
        s3_credentials
        delimiter
        do_build_client

        Returns
        -------

        """

        self.aws_bucket_credentials = s3_credentials
        self.files: List[Any] = []
        self.delimiter = delimiter
        self._current_pos: int = -1
        self._client: Any = None

        if do_build_client:
            self.build_client()

    def build_client(self) -> Any:
        """Build an S3 client from the given AWS S3 credentials

        Returns
        -------

        """
        self._client = get_aws_s3_client(credentials=self.aws_bucket_credentials)

    def load_from_list(self, files: List[str], delimiter="/"):
        self.files = files
        self.delimiter = delimiter

    def read_file_byte_string(
        self,
        key: Union[int, str],
        read_from_local_host: bool = False,
        file_reader: Callable = None,
    ) -> str:
        """Returns the byte string associated with the given key

        Parameters
        ----------
        file_reader
        read_from_local_host
        key: The key of the object to read

        Returns
        -------

        The byte string of the object that is represented by the
        given key

        """

        if read_from_local_host:
            image = self[key]
            return file_reader(image)

        file_obj = self.get_object(key=key)

        if "Body" not in file_obj:
            raise ValueError("Body is missing from file object response")

        return file_obj["Body"].read()

    def read(
        self, prefixes: tuple, valid_image_extensions: tuple, delimiter: str = "/"
    ):
        self.delimiter = delimiter

        # empty the files as we will read new ones
        self.files = []

        if self._client is None:
            self.build_client()

        # get the object that lists the objects
        # on S3
        paginator = self._client.get_paginator("list_objects")

        for prefix in prefixes:
            s3_iterator = paginator.paginate(
                Bucket=self.aws_bucket_credentials.aws_s3_bucket_name,
                Delimiter=self.delimiter,
                Prefix=prefix,
            )

            contents = expand_s3_iterator_contents(s3_iterator)

            if len(contents) != 0:
                self.read_from_contents(
                    contents=contents, valid_image_extensions=valid_image_extensions
                )

            common_prefixes = expand_s3_iterator_common_prefixes(
                s3_iterator=s3_iterator
            )
            if len(common_prefixes) != 0:
                self.read_from_common_prefixes(
                    common_prefixes=common_prefixes,
                    valid_image_extensions=valid_image_extensions,
                )

    def copy_to(self, s3_credentials_to: AWSCredentials_S3) -> None:
        """Copy the images in the bucket to the S3 bucket specified
        by the provided credentials

        Parameters
        ----------
        s3_credentials_to: The credentials to use for the bucket to copy

        Returns
        -------

        """

        if len(self.files) == 0:
            print("WARNING: Image batch is empty...")
            return

        raise NotImplementedError("The function is not implemented")

    def copy_file_to(
        self,
        key: Union[int, str],
        s3_credentials_to: AWSCredentials_S3,
        new_filename: str = "",
        **kwargs,
    ) -> dict:
        """Copies the specified file identified by the given key
        into the bucket specified in the given AWSCredentials_S3 credentials

        Parameters
        ----------
        key: The index of the file to copy to
        s3_credentials_to: Credentials for accessing the new bucket
        kwargs: Arguments to orchestrate how the copy is done

        Returns
        -------

        """

        item_to_copy = self[key]

        if new_filename == "":
            new_filename = item_to_copy

        if self._client is None:
            self.build_client()

        copy_source = (
            self.aws_bucket_credentials.aws_s3_bucket_name + "/" + item_to_copy
        )
        copy_response = self._client.copy_object(
            Bucket=s3_credentials_to.aws_s3_bucket_name,
            Key=new_filename,
            CopySource=copy_source,
            **kwargs,
        )
        return copy_response

    def get_object(self, key: Union[int, str]) -> Any:
        """Returns the object specified with the given key.

        Parameters
        ----------
        key: The key of the object to return

        Returns
        -------

        """

        if type(key) == int:
            if key >= len(self.files):
                raise ValueError(
                    f"Invalid key. Integer key {key} cannot be >= {len(self.files)}"
                )

            key = self.files[key]
        elif type(key) == str:
            if key not in self.files:
                raise ValueError(f"key={key} not in {self.files}")
        else:
            raise ValueError(f"key type {type(key)} is not valid")

        if self._client is None:
            raise ValueError("S3 client is not built")

        file_byte_string = self._client.get_object(
            Bucket=self.aws_bucket_credentials.aws_s3_bucket_name, Key=key
        )

        return file_byte_string

    def read_from_contents(
        self, contents: List[dict], valid_image_extensions: tuple
    ) -> None:
        """Read the contents of the S3 bucket

        Parameters
        ----------
        contents: The contents to use for reading the images
        valid_image_extensions: The valid_image_extensions to look for

        Returns
        -------

        """

        n_extensions = len(valid_image_extensions)
        for content in contents:
            image = content.get("Key")

            if n_extensions != 0:
                img_extension = os.path.splitext(image)[1]
                if img_extension in valid_image_extensions:
                    self.files.append(content.get("Key"))
            else:
                self.files.append(content.get("Key"))

    def read_from_common_prefixes(
        self, common_prefixes: List[dict], valid_image_extensions: tuple
    ) -> None:
        """Read from the common_prefixes

        Parameters
        ----------
        common_prefixes: The common_prefixes to use for reading
        valid_image_extensions: The valid_image_extensions to look for

        Returns
        -------

        """

        if self._client is None:
            raise ValueError("S3 client is not built")

        paginator = self._client.get_paginator("list_objects")
        for item in common_prefixes:
            prefix = item["Prefix"]

            s3_iterator = paginator.paginate(
                Bucket=self.aws_bucket_credentials.aws_s3_bucket_name,
                Delimiter=self.delimiter,
                Prefix=prefix,
            )

            contents = expand_s3_iterator_contents(s3_iterator)
            self.read_from_contents(
                contents=contents, valid_image_extensions=valid_image_extensions
            )



from functools import wraps
from typing import Callable


def async_aws_lambda_task_decorator(
    async_before_task_runner: Callable, async_after_task_runner: Callable
):
    def _execute_lambda_task(async_lambda_task_fn: Callable):
        @wraps(async_lambda_task_fn)
        async def wrapper(*args, **kwargs):
            if async_before_task_runner is not None:
                before_task_runner_result = await async_before_task_runner(
                    *args, **kwargs
                )

            lambda_task_result = await async_lambda_task_fn(*args, **kwargs)

            if async_after_task_runner is not None:
                after_task_runner_result = await async_after_task_runner(
                    *args, **kwargs
                )
            return lambda_task_result

        return wrapper

    return _execute_lambda_task


def aws_lambda_task_decorator(
    before_task_runner: Callable, after_task_runner: Callable
):
    def _execute_lambda_task(lambda_task_fn: Callable):
        @wraps(lambda_task_fn)
        def wrapper(*args, **kwargs):
            if before_task_runner is not None:
                before_task_runner_result = before_task_runner(*args, **kwargs)
            lambda_task_result = lambda_task_fn(*args, **kwargs)

            if after_task_runner is not None:
                after_task_runner_result = after_task_runner(*args, **kwargs)
            return lambda_task_result

        return wrapper

    return _execute_lambda_task



from typing import List, Union


class MessageAttributes:
    def __init__(self):
        self.string_value: str = ""
        self.binary_value: bytes = b"bytes"
        self.string_list_values: List[str]
        self.binary_list_values: List[b"bytes"]
        self.data_type: str


class SQSMessageConfig(object):
    def __init__(
        self,
        message_body: str,
        message_group_id: str,
        message_attributes: Union[MessageAttributes, None],
        message_deduplication_id: str,
        delay_seconds: int = 123,
    ):
        self.message_body: str = message_body
        self.message_group_id = message_group_id
        self.message_deduplication_id = message_deduplication_id
        self.message_attributes = message_attributes
        self.delay_seconds = delay_seconds



from navalmartin_mir_aws_utils.aws_credentials import AWSCredentials_SQS
from navalmartin_mir_aws_utils.sqs_queue_config import SQSMessageConfig
from navalmartin_mir_aws_utils.sqs_utils import (
    send_sqs_message,
    read_one_sqs_message,
    delete_sqs_message,
)

AWS_REGION = "YOUR_AWS_REGION"
AWS_S3_BUCKET_NAME = "YOUR_AWS_S3_BUCKET_NAME"
AWS_SQS_URL = "YOUR_AWS_SQS_URL"
AWS_SQS_NAME = "YOUR_AWS_SQS_NAME"
AWS_SQS_GROUP_ID = "YOUR_AWS_SQS_GROUP_ID"


if __name__ == "__main__":
    aws_sqs_credentials = AWSCredentials_SQS(
        aws_sqs_queue_name=AWS_SQS_NAME,
        aws_queue_url=AWS_SQS_URL,
        aws_region=AWS_REGION,
    )

    message = "task_id:123"
    sqs_msg = SQSMessageConfig(
        message_body=message,
        message_group_id=AWS_SQS_GROUP_ID,
        message_attributes=None,
        message_deduplication_id="123",
    )

    response = send_sqs_message(sqs_credentials=aws_sqs_credentials, sqs_msg=sqs_msg)

    print(response)

    read_response = read_one_sqs_message(aws_sqs_credentials)
    print(read_response)

    receipt_handle = read_response["Messages"][0]["ReceiptHandle"]
    # get the receipt handler to delete the message
    # this signals that the message has been consumed
    delete_response = delete_sqs_message(
        aws_sqs_credentials, receipt_handle=receipt_handle
    )
    print(delete_response)



import base64
import hmac
import hashlib


def get_secret_hash(username: str, client_id: str, client_secret: str) -> str:
    """Calculate a secret hash  for the given username on the
    given client id
    It uses the SHA256 algorithm

    Parameters
    ----------
    username: The username
    client_id: The client id
    client_secret: The client secret

    Returns
    -------

    A string representing the hash
    """
    msg = username + client_id
    dig = hmac.new(
        str(client_secret).encode("utf-8"),
        msg=str(msg).encode("utf-8"),
        digestmod=hashlib.sha256,
    ).digest()
    d2 = base64.b64encode(dig).decode()
    return d2



from navalmartin_mir_aws_utils.aws_credentials import AWSCredentials_CognitoIDP
from navalmartin_mir_aws_utils.cognito_idp_utils import (
    global_signout_user_from_pool,
    authenticate_and_get_token_for_user,
)
from navalmartin_mir_aws_utils.utils import AWSCognitoSignInUserData

AWS_REGION = ""
AWS_COGNITO_POOL_ID = ""
AWS_COGNITO_CLIENT_ID = ""
AWS_COGNITO_CLIENT_SECRET = ""

if __name__ == "__main__":
    user_data = AWSCognitoSignInUserData(username="", password="")

    aws_credentials = AWSCredentials_CognitoIDP(
        aws_region=AWS_REGION,
        aws_cognito_pool_id=AWS_COGNITO_POOL_ID,
        aws_cognito_client_id=AWS_COGNITO_CLIENT_ID,
        aws_cognito_client_secret=AWS_COGNITO_CLIENT_SECRET,
        aws_access_key=None,
        aws_secret_access_key=None,
    )

    response = authenticate_and_get_token_for_user(
        aws_cognito_credentials=aws_credentials, user_data=user_data
    )

    access_token = response["AuthenticationResult"]["AccessToken"]
    print(f"Access token {access_token}")

    response = global_signout_user_from_pool(
        access_token=access_token, credentials=aws_credentials
    )

    print(f"Revoke response {response}")



"""module simple_email. Models an email to be sent via SES
"""
from typing import List, Optional, Dict
from pydantic import EmailStr, BaseModel, Field


class EmailBody(BaseModel):
    data: str = Field(
        title="data", description="The actual email content we want to send"
    )
    charset: Optional[str] = Field(
        title="charset", description="The character set used in the data"
    )
    html: Optional[Dict] = Field(
        title="html",
        description="The content of the message, in HTML format. Use this for email "
        "clients that can process HTML. You can include clickable links, "
        "formatted text, and much more in an HTML message.",
    )


class EmailSubject(BaseModel):
    data: str = Field(title="data", description="The actual subject of the email")
    charset: Optional[str] = Field(
        title="charset", description="The character set used in the data"
    )


class Email(BaseModel):
    source: EmailStr = Field(title="source", description="The source email")
    destination: EmailStr = Field(title="destination", description="The destination email")
    subject: EmailSubject = Field(title="subject")
    body: EmailBody = Field(title="body")
    ccs: Optional[List[EmailStr]] = Field(title="ccs", default=[])
    reply_to: Optional[EmailStr] = Field(title="reply_to", default=None)
    charset: Optional[str] = Field(title="charset",
                                   description="The character set to use in the email",
                                   default="UTF-8")



"""module aws_credentials. Utility class to move around
credentials. This class is not used to load the
credentials. It is the application's duty to do so

"""


class AWSCredentials(object):
    def __init__(
            self,
            aws_client_name: str = None,
            aws_access_key: str = None,
            aws_secret_access_key: str = None,
    ):
        self.aws_client_name: str = aws_client_name
        self.aws_access_key: str = aws_access_key
        self.aws_secret_access_key = aws_secret_access_key


class AWSCredentials_S3(AWSCredentials):
    def __init__(
            self,
            aws_s3_bucket_name: str,
            aws_region: str,
            aws_access_key: str = None,
            aws_secret_access_key: str = None,
    ):
        super(AWSCredentials_S3, self).__init__(
            aws_client_name="s3",
            aws_secret_access_key=aws_secret_access_key,
            aws_access_key=aws_access_key,
        )
        self.aws_s3_bucket_name = aws_s3_bucket_name
        self.aws_region = aws_region


class AWSCredentials_SQS(AWSCredentials):
    def __init__(
            self,
            aws_sqs_queue_name: str,
            aws_region: str,
            aws_queue_url: str,
            aws_access_key: str = None,
            aws_secret_access_key: str = None,
    ):
        super(AWSCredentials_SQS, self).__init__(
            aws_client_name="sqs",
            aws_secret_access_key=aws_secret_access_key,
            aws_access_key=aws_access_key,
        )
        self.aws_sqs_queue_name: str = aws_sqs_queue_name
        self.aws_region: str = aws_region
        self.queue_url: str = aws_queue_url


class AWSCredentials_CognitoIDP(AWSCredentials):
    def __init__(
            self,
            aws_region: str,
            aws_cognito_pool_id: str,
            aws_cognito_client_id: str,
            aws_cognito_client_secret: str,
            aws_access_key: str = None,
            aws_secret_access_key: str = None,
    ):
        super(AWSCredentials_CognitoIDP, self).__init__(
            aws_client_name="cognito-idp",
            aws_access_key=aws_access_key,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.aws_region = aws_region
        self.aws_cognito_pool_id = aws_cognito_pool_id
        self.aws_cognito_client_id = aws_cognito_client_id
        self.aws_cognito_client_secret = aws_cognito_client_secret


class AWSCredentials_SecretsManager(AWSCredentials):
    def __init__(
            self,
            aws_region: str,
            secret_name: str,
            aws_access_key: str = None,
            aws_secret_access_key: str = None,
    ):
        super(AWSCredentials_SecretsManager, self).__init__(
            aws_client_name="secretsmanager",
            aws_access_key=aws_access_key,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.aws_region = aws_region
        self.secret_name = secret_name


class AWSCredentials_SES(AWSCredentials):
    def __init__(
            self,
            aws_region: str,
            aws_access_key: str = None,
            aws_secret_access_key: str = None,
    ):
        super(AWSCredentials_SES, self).__init__(
            aws_client_name="ses",
            aws_access_key=aws_access_key,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.aws_region = aws_region


class AWSCredentials_SFN(AWSCredentials):
    def __init__(
            self,
            state_machine_arn: str,
            aws_region: str,
            aws_access_key: str = None,
            aws_secret_access_key: str = None,
    ):
        super(AWSCredentials_SFN, self).__init__(
            aws_client_name="sfn",
            aws_access_key=aws_access_key,
            aws_secret_access_key=aws_secret_access_key,
        )

        self.aws_region = aws_region
        self.state_machine_arn = state_machine_arn


class AWSCredentials_SageMaker(AWSCredentials):
    def __init__(self, aws_region: str,
                 aws_access_key: str = None,
                 aws_secret_access_key: str = None):
        super(AWSCredentials_SageMaker, self).__init__(aws_client_name='sagemaker',
                                                       aws_access_key=aws_access_key,
                                                       aws_secret_access_key=aws_secret_access_key)
        self.aws_region = aws_region



from typing import Any, List, ByteString
from navalmartin_mir_aws_utils.aws_credentials import AWSCredentials_S3
from navalmartin_mir_aws_utils.boto3_client import get_aws_s3_client


def get_s3_iterator(
        s3_client: Any, prefix: str, aws_creds: AWSCredentials_S3, delimiter="/"
) -> Any:
    paginator = s3_client.get_paginator("list_objects")
    return paginator.paginate(
        Bucket=aws_creds.aws_s3_bucket_name, Delimiter=delimiter, Prefix=prefix
    )


def expand_s3_iterator_contents(s3_iterator: Any) -> List[dict]:
    """Get the contents of the provided S3 iterator

    Parameters
    ----------
    s3_iterator: The S3 iterator

    Returns
    -------
    A list of dictionaries
    """

    contents = s3_iterator.search("Contents")
    return [item for item in contents if item is not None]


def expand_s3_iterator_common_prefixes(s3_iterator: Any) -> List[dict]:
    """Get the common prefixes from the iterator

    Parameters
    ----------
    s3_iterator: The S3 iterator

    Returns
    -------
    A list of dictionaries
    """
    contents = s3_iterator.search("CommonPrefixes")
    return [item for item in contents if item is not None]


def delete_s3_all_objs_with_key(
        keys: List[dict], s3_client: Any, aws_creds: AWSCredentials_S3, **options
) -> dict:
    """Delete all the objects on S3 with the given key.

    Parameters
    ----------
    keys: The prefix of the objects
    s3_client: The S3 client the objects sit on
    aws_creds: The credentials to use to create an
    S3 client: if the user provided if None an s3 client is built
    options: Options to pass in the boto3 API call. Currently, the following
    options are used quiet: True | False

    Returns
    -------

    A dictionary with the response of the operation
    """

    if keys is None or len(keys) == 0:
        raise ValueError("Invalid key given")

    if aws_creds is None:
        raise ValueError("AWS credentials not given")

    if s3_client is None:
        s3_client = get_aws_s3_client(credentials=aws_creds)

    delete_dict = {"Objects": keys, "Quiet": True if "quiet" in options else False}
    response = s3_client.delete_objects(
        Bucket=aws_creds.aws_s3_bucket_name, Delete=delete_dict
    )

    return response


def delete_s3_object_with_key(
        key: str, s3_client: Any, aws_creds: AWSCredentials_S3, **options
) -> dict:
    if key is None or key == "":
        raise ValueError("Invalid key given")

    if aws_creds is None:
        raise ValueError("AWS credentials not given")

    if s3_client is None:
        s3_client = get_aws_s3_client(credentials=aws_creds)

    if "version_id" in options:
        version_id = options["version_id"]
        response = s3_client.delete_object(
            Bucket=aws_creds.aws_s3_bucket_name, Key=key, VersionId=version_id
        )
    else:
        response = s3_client.delete_object(Bucket=aws_creds.aws_s3_bucket_name, Key=key)

    return response


def save_object_to_s3(
        body: ByteString, key: str, s3_client: Any, aws_creds: AWSCredentials_S3
) -> dict:
    """
    Save object to designated S3 bucket
    Parameters
    ----------
    body: The object in BytesString
    key: The prefix of the object
    s3_client: The S3 client the objects sit on
    aws_creds: The aws credentials to use

    Returns
    -------
    A dictionary with the response
    """
    if key is None or key == "":
        raise ValueError("Invalid key given")

    if aws_creds is None:
        raise ValueError("AWS credentials not given")

    if s3_client is None:
        s3_client = get_aws_s3_client(credentials=aws_creds)

    response = s3_client.put_object(
        Body=body, Bucket=aws_creds.aws_s3_bucket_name, Key=key
    )

    return response


def generate_presigned_url(object_name: str,
                           aws_creds: AWSCredentials_S3,
                           client_method_name: str = 'get_object',
                           expiration: int = 3600,
                           s3_client: Any = None) -> dict:
    """Generates a signed url for AWS S3

    Parameters
    ----------
    client_method_name: Name of the S3.Client method, e.g., 'list_buckets'
    expiration: Time in seconds for the presigned URL to remain valid
    object_name: The item in AWS S3 to access
    aws_creds: AWS credentials to sign in
    s3_client

    Returns
    -------

    A dictionary with the response
    """

    if s3_client is None:
        if aws_creds is None:
            raise ValueError("aws_creds and s3_client parameters cannot be simultaneously None")
        else:
            s3_client = get_aws_s3_client(credentials=aws_creds)

    response = s3_client.generate_presigned_url(client_method_name,
                                                Params={'Bucket': aws_creds.aws_s3_bucket_name,
                                                        'Key': object_name},
                                                ExpiresIn=expiration)
    return response



from pydantic import EmailStr
from navalmartin_mir_aws_utils import AWSCredentials_SES
from navalmartin_mir_aws_utils.tasks import send_simple_email, send_html_email
from navalmartin_mir_aws_utils.utils import Email, EmailSubject, EmailBody


if __name__ == "__main__":
    credentials = AWSCredentials_SES(aws_region="eu-west-2")
    email = Email(
        source=EmailStr("mir-surveys@mir.blue"),
        destination=EmailStr("alex@navalmartin.com"),
        subject=EmailSubject(data="First notification from mir"),
        body=EmailBody(
            data="This your first AWS SES notification from mir. Expect more."
        ),
    )

    print(email.dict())
    response = send_simple_email(email=email, credentials=credentials)
    print(response)

    email_html = """<html>
    <p>Your survey PDF has been compiled. You can access it here <a href="{0}">my-survey</a>.
    Alternatively, you can login to your account <a href="https://mir.blue/session/login">here</a>.
    </p>
    
    <p>
    For any questions or clarifications you may have regarding this survey, please get in 
    touch at mir-surveys@min.blue. Please use the survey id in your email.</>
    
    <p>Thank you for using mir.</p>
    
    </html>
    """.format("https://mir.blue/session/login")
    email = Email(
        source=EmailStr("mir-surveys@mir.blue"),
        destination=EmailStr("alex@navalmartin.com"),
        subject=EmailSubject(data="Your survey 1235 is ready."),
        body=EmailBody(
            data=email_html
        ),
    )

    print(email.dict())
    response = send_html_email(email=email, credentials=credentials)
    print(response)
